@@ -16,6 +16,8 @@
 # along with this program; If not, see <http://www.gnu.org/licenses/>.
 
 import logging
+from typing import Any
+from typing import Optional
 
 from gi.repository import GLib
 
@@ -27,15 +29,17 @@ from nbxmpp.protocol import BindRequest
 from nbxmpp.protocol import TLSRequest
 from nbxmpp.protocol import isResultNode
 from nbxmpp.protocol import JID
-from nbxmpp.protocol import Iq
 from nbxmpp.protocol import Protocol
 from nbxmpp.protocol import WebsocketCloseHeader
+from nbxmpp.errors import TimeoutStanzaError
+from nbxmpp.errors import StanzaError
+from nbxmpp.errors import CancelledError
 from nbxmpp.addresses import ServerAddresses
 from nbxmpp.addresses import NoMoreAddresses
 from nbxmpp.tcp import TCPConnection
 from nbxmpp.websocket import WebsocketConnection
 from nbxmpp.smacks import Smacks
-from nbxmpp.auth import SASL
+from nbxmpp.sasl import SASL
 from nbxmpp.const import StreamState
 from nbxmpp.const import StreamError
 from nbxmpp.const import ConnectionType
@@ -43,10 +47,9 @@ from nbxmpp.const import ConnectionProtocol
 from nbxmpp.const import Mode
 from nbxmpp.dispatcher import StanzaDispatcher
 from nbxmpp.util import get_stream_header
-from nbxmpp.util import get_stanza_id
+from nbxmpp.util import generate_id
 from nbxmpp.util import Observable
 from nbxmpp.util import validate_stream_header
-from nbxmpp.util import is_error_result
 from nbxmpp.util import LogAdapter
 
 log = logging.getLogger('nbxmpp.stream')
@@ -95,6 +98,8 @@ class Client(Observable):
         self._allowed_protocols = None
         self._allowed_mechs = None
 
+        self._sm_disabled = False
+
         self._stream_id = None
         self._stream_secure = False
         self._stream_authenticated = False
@@ -102,10 +107,10 @@ class Client(Observable):
         self._session_required = False
         self._connect_successful = False
         self._stream_close_initiated = False
-        self._ping_id = None
+        self._ping_task = None
         self._error = None, None, None
 
-        self._ignored_tls_errors = []
+        self._ignored_tls_errors = set()
         self._ignore_tls_errors = False
         self._accepted_certificates = []
         self._peer_certificate = None
@@ -113,9 +118,11 @@ class Client(Observable):
 
         self._con = None
         self._local_address = None
+        self._remote_address = None
         self._mode = Mode.CLIENT
 
         self._ping_source_id = None
+        self._tasks = []
 
         self._dispatcher = StanzaDispatcher(self)
         self._dispatcher.subscribe('before-dispatch', self._on_before_dispatch)
@@ -127,6 +134,15 @@ class Client(Observable):
 
         self._state = StreamState.DISCONNECTED
 
+    def add_task(self, task):
+        self._tasks.append(task)
+
+    def remove_task(self, task, _context):
+        try:
+            self._tasks.remove(task)
+        except Exception:
+            pass
+
     @property
     def log_context(self):
         return self._log_context
@@ -135,6 +151,10 @@ class Client(Observable):
     def features(self):
         return self._stream_features
 
+    @property
+    def resumeable(self):
+        return self._smacks.resumeable
+
     @property
     def sm_supported(self):
         return self._smacks.sm_supported
@@ -184,6 +204,8 @@ class Client(Observable):
         return self._ignored_tls_errors
 
     def set_ignored_tls_errors(self, errors):
+        if errors is None:
+            errors = set()
         self._ignored_tls_errors = errors
 
     @property
@@ -204,6 +226,10 @@ class Client(Observable):
     def peer_certificate(self):
         return self._peer_certificate, self._peer_certificate_errors
 
+    @property
+    def current_address(self):
+        return self._current_address
+
     @property
     def current_connection_type(self):
         return self._current_address.type
@@ -242,6 +268,10 @@ class Client(Observable):
     def local_address(self):
         return self._local_address
 
+    @property
+    def remote_address(self):
+        return self._remote_address
+
     @property
     def connection_types(self):
         if self._custom_host is not None:
@@ -254,10 +284,9 @@ class Client(Observable):
 
     @property
     def mechs(self):
-        return set(self._allowed_mechs or set(['SCRAM-SHA-256',
+        return set(self._allowed_mechs or set(['SCRAM-SHA-512',
+                                               'SCRAM-SHA-256',
                                                'SCRAM-SHA-1',
-                                               # 'SCRAM-SHA-256-PLUS',
-                                               # 'SCRAM-SHA-1-PLUS',
                                                'PLAIN']))
 
     def set_mechs(self, mechs):
@@ -273,6 +302,13 @@ class Client(Observable):
     def set_protocols(self, protocols):
         self._allowed_protocols = protocols
 
+    def set_sm_disabled(self, value):
+        self._sm_disabled = value
+
+    @property
+    def sm_disabled(self):
+        return self._sm_disabled
+
     @property
     def client_cert(self):
         return self._client_cert, self._client_cert_pass
@@ -293,7 +329,7 @@ class Client(Observable):
         return self._jid
 
     def _set_bound_jid(self, jid):
-        self._jid = JID(jid)
+        self._jid = JID.from_string(jid)
 
     @property
     def has_error(self):
@@ -407,7 +443,8 @@ class Client(Observable):
     def _disconnect(self, immediate=True):
         self.state = StreamState.DISCONNECTING
         self._remove_ping_timer()
-        self._dispatcher.remove_ping_callback(self._ping_id)
+        self._cancel_ping_task()
+
         if not immediate:
             self._stream_close_initiated = True
             self._smacks.close_session()
@@ -423,11 +460,14 @@ class Client(Observable):
     def _on_connected(self, connection, _signal_name):
         self.set_state(StreamState.CONNECTED)
         self._local_address = connection.local_address
+        self._remote_address = connection.remote_address
 
     def _on_disconnected(self, _connection, _signal_name):
         self.state = StreamState.DISCONNECTED
+        for task in self._tasks:
+            task.cancel()
         self._remove_ping_timer()
-        self._dispatcher.remove_ping_callback(self._ping_id)
+        self._cancel_ping_task()
         self._reset_stream()
         self.notify('disconnected')
 
@@ -439,8 +479,8 @@ class Client(Observable):
         else:
             self._set_error(StreamError.CONNECTION_FAILED,
                             'connection-failed',
-                            (f'Unable to connect to last '
-                             'successful address: {self._current_address}'))
+                            ('Unable to connect to last '
+                             f'successful address: {self._current_address}'))
             self.notify('connection-failed')
 
     def _disconnect_with_error(self, error_domain, error, text=None):
@@ -461,7 +501,7 @@ class Client(Observable):
         if not self._stream_close_initiated:
             self.state = StreamState.DISCONNECTING
             self._remove_ping_timer()
-            self._dispatcher.remove_ping_callback(self._ping_id)
+            self._cancel_ping_task()
             self._smacks.close_session()
             self._end_stream()
             self._con.shutdown_output()
@@ -533,8 +573,13 @@ class Client(Observable):
         GLib.source_remove(self._ping_source_id)
         self._ping_source_id = None
 
-    def send_stanza(self, stanza, now=False, callback=None,
-                    timeout=None, user_data=None):
+    def send_stanza(self,
+                    stanza: Protocol,
+                    now: bool = False,
+                    callback: Optional[Any] = None,
+                    timeout: Optional[int] = None,
+                    user_data: Optional[Any] = None) -> str:
+
         if user_data is not None and not isinstance(user_data, dict):
             raise ValueError('arg user_data must be of dict type')
 
@@ -543,7 +588,7 @@ class Client(Observable):
 
         id_ = stanza.getID()
         if id_ is None:
-            id_ = get_stanza_id()
+            id_ = generate_id()
             stanza.setID(id_)
 
         if callback is not None:
@@ -600,6 +645,11 @@ class Client(Observable):
                 # other connection methods if an error happensafterwards
                 self._connect_successful = True
 
+            if self._stream_authenticated and self._mode.is_login_test:
+                self.notify('login-successful')
+                self.disconnect()
+                return
+
             self.state = StreamState.WAIT_FOR_FEATURES
 
         elif self.state == StreamState.WAIT_FOR_FEATURES:
@@ -642,15 +692,10 @@ class Client(Observable):
 
         elif self.state == StreamState.AUTH_SUCCESSFUL:
             self._stream_authenticated = True
-            if self._mode.is_login_test:
-                self.notify('login-successful')
-                # Reset parser because we will receive a new stream header
-                # which will otherwise lead to a parsing error
-                self._dispatcher.reset_parser()
-                self.disconnect()
-                return
-
-            self._start_stream()
+            if self._sasl.is_sasl2():
+                self.state = StreamState.WAIT_FOR_FEATURES
+            else:
+                self._start_stream()
 
         elif self.state == StreamState.AUTH_FAILED:
             self._disconnect_with_error(StreamError.SASL,
@@ -744,7 +789,7 @@ class Client(Observable):
         self.state = StreamState.WAIT_FOR_TLS_PROCEED
 
     def _start_auth(self, features):
-        if not features.has_sasl():
+        if not features.has_sasl() and not features.has_sasl_2():
             self._log.error('Server does not support SASL')
             self._disconnect_with_error(StreamError.SASL,
                                         'sasl-not-supported')
@@ -790,21 +835,33 @@ class Client(Observable):
 
     def _ping(self):
         self._ping_source_id = None
-        iq = Iq('get', to=self.domain)
-        iq.addChild(name='ping', namespace=Namespace.PING)
-        self._ping_id = self.send_stanza(iq,
-                                         timeout=10,
-                                         callback=self._on_pong)
-        self._log.info('Ping')
+        self._ping_task = self.get_module('Ping').ping(
+            self.domain,
+            timeout=10,
+            callback=self._on_pong)
+
+    def _on_pong(self, task):
+        self._ping_task = None
+
+        try:
+            task.finish()
+        except TimeoutStanzaError:
+            self._log.info('Ping timeout')
+            self._disconnect(immediate=True)
+            return
+
+        except CancelledError:
+            return
+
+        except StanzaError:
+            pass
 
-    def _on_pong(self, _client, result):
-        if is_error_result(result):
-            if result.condition == 'timeout':
-                self._log.info('Ping timeout')
-                self._disconnect(immediate=True)
-                return
         self._log.info('Pong')
 
+    def _cancel_ping_task(self):
+        if self._ping_task is not None:
+            self._ping_task.cancel()
+
     def register_handler(self, *args, **kwargs):
         self._dispatcher.register_handler(*args, **kwargs)
 
@@ -812,6 +869,8 @@ class Client(Observable):
         self._dispatcher.unregister_handler(*args, **kwargs)
 
     def destroy(self):
+        for task in self._tasks:
+            task.cancel()
         self._remove_ping_timer()
         self._smacks = None
         self._sasl = None
