@@ -44,6 +44,7 @@ def ascii_upper(s):
     return s.upper()
 
 SASL_AUTH_MECHS = [
+    'SCRAM-SHA-512-PLUS',
     'SCRAM-SHA-512',
     'SCRAM-SHA-256-PLUS',
     'SCRAM-SHA-256',
@@ -520,7 +521,7 @@ def validate_domainpart(domainpart: str) -> str:
         domainpart = domainpart[:-1]
 
     try:
-        idna_encode(domainpart)
+        domainpart = idna2008_prep(domainpart)
     except Exception:
         raise DomainpartNotAllowedChar
 
@@ -528,8 +529,18 @@ def validate_domainpart(domainpart: str) -> str:
 
 
 @functools.lru_cache(maxsize=None)
-def idna_encode(domain: str) -> str:
-    return idna.encode(domain, uts46=True).decode()
+def idna2008_prep(domain: str, to_ascii: bool = False) -> str:
+    '''
+    Prepare with UTS46 case mapping to stay compatibel with the IDNA2003
+    mapping. Further try to encode the domain to catch illegal domains.
+    Only return the case mapped domain because on the XMPP wire,UTF8 domains
+    are fine.
+    '''
+    domain = idna.uts46_remap(domain)
+    encoded_domain = idna.encode(domain)
+    if to_ascii:
+        return encoded_domain.decode()
+    return domain
 
 
 @functools.lru_cache(maxsize=None)
@@ -685,7 +696,7 @@ class JID:
         return not self.__eq__(other)
 
     def domain_to_ascii(self) -> str:
-        return idna_encode(self.domain)
+        return idna2008_prep(self.domain, to_ascii=True)
 
     @property
     def bare(self) -> str:
@@ -700,9 +711,7 @@ class JID:
     def new_as_bare(self) -> JID:
         if self.resource is None:
             return self
-        new = asdict(self)
-        new.pop('resource')
-        return JID(**new)
+        return JID.from_string(self.bare)
 
     def bare_match(self, other: Union[str, JID]) -> bool:
         if isinstance(other, str):
@@ -1710,10 +1719,15 @@ class Features(Node):
     def has_sasl(self):
         return self.getTag('mechanisms', namespace=Namespace.SASL) is not None
 
+    def has_sasl_2(self):
+        return self.getTag('mechanisms', namespace=Namespace.SASL2) is not None
+
     def get_mechs(self) -> set[str]:
-        mechanisms = self.getTag('mechanisms', namespace=Namespace.SASL)
+        mechanisms = self.getTag('mechanisms', namespace=Namespace.SASL2)
         if mechanisms is None:
-            return set()
+            mechanisms = self.getTag('mechanisms', namespace=Namespace.SASL)
+            if mechanisms is None:
+                return set()
         mechanisms = mechanisms.getTags('mechanism')
         return set(mech.getData() for mech in mechanisms)
 
